/* smpw.c - self-maintaining indexed access to /etc/passwd
**
** Copyright (C) 1992 by Jef Poskanzer.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.
**
**
** Parts of this file are:
** Copyright 1989, 1990, 1991 John F. Haugh II
** All rights reserved.
**
** Permission is granted to copy and create derivative works for any
** non-commercial purpose, provided this copyright notice is preserved
** in all copies of source code, or included in human readable form
** and conspicuously displayed on all copies of object code or
** distribution media.
*/

#include <stdio.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>

#ifdef SYSV
#include <string.h>
#define	index strchr
#else /*SYSV*/
#include <strings.h>
#endif /*SYSV*/

extern off_t lseek();

#ifndef PWDFILE
#define PWDFILE "/etc/passwd"
#endif
#ifndef PWDUFILE
#define PWDUFILE "/etc/passwd.u"
#endif
#ifndef PWDNFILE
#define PWDNFILE "/etc/passwd.n"
#endif
#ifndef PWDSFILE
#define PWDSFILE "/etc/passwd.s"
#endif

#define NAMEHASHMULT 50021

#define NO_OFF 2147483647
#define NO_UID 65535

#define NEARBY_ENTRIES 30
#define BYTES_PER_ENTRY 74	/* approximately */

#define NFIELDS 7

static uid_t maxuid;
static off_t namehashsize;

static FILE* pwdfp = (FILE*) 0;
static int pwufd = -1;
static int pwnfd = -1;

static char pwdbuf[BUFSIZ];
static char* pwdfields[NFIELDS];
static struct passwd pwent;

/* smpw_init - initialize the library */
static void
smpw_init()
    {
    static int firstTime = 1;
    FILE* fp;
    int i1, i2;

    if ( firstTime )
	{
	firstTime = 0;

	/* Read maxuid and namehashsize from the file.  The reason we do
	** this instead of just compiling them in is so that if you ever
	** need to change them, you don't have to recompile every program
	** in the world.
	*/
	maxuid = 0;
	fp = fopen( PWDSFILE, "r" );
	if ( fp != (FILE*) 0 )
	    {
	    if ( fscanf( fp, "%d %d", &i1, &i2 ) == 2 )
		{
		maxuid = i1;
		namehashsize = i2;
		}
	    (void) fclose( fp );
	    }
	}
    }

/* sgetpwent - convert a string to a (struct passwd)
**
** sgetpwent() parses pwdbuf into the parts required for a password
** structure.  Strict checking is made for the uid and gid fields and
** presence of the correct number of colons.  Any failing tests result
** in a NULL pointer being returned.
**/
static struct passwd*
sgetpwent()
    {
    int i;
    char* cp;

    /* Save a pointer to the start of each colon separated
    ** field.  The fields are converted into NUL terminated strings.
    */
    for ( cp = pwdbuf, i = 0; i < NFIELDS && cp; ++i )
	{
	pwdfields[i] = cp;
	cp = index( cp, ':' );
	if ( cp )
	    *cp++ = 0;
	}

    /* There must be exactly NFIELDS colon separated fields or
    ** the entry is invalid.  Also, the uid and gid must be non-blank.
    */
    if ( i != NFIELDS || *pwdfields[2] == '\0' || *pwdfields[3] == '\0' )
	return (struct passwd*) 0;

    /* Each of the fields is converted the appropriate data type
    ** and the result assigned to the password structure.  If the
    ** uid or gid does not convert to an integer value, a NULL
    ** pointer is returned.
    */
    pwent.pw_name = pwdfields[0];
    pwent.pw_passwd = pwdfields[1];
    if ( sscanf( pwdfields[2], "%d", &i ) != 1 )
	return (struct passwd*) 0;
    pwent.pw_uid = i;
    if ( sscanf( pwdfields[3], "%d", &i ) != 1 )
	return (struct passwd*) 0;
    pwent.pw_gid = i;
    pwent.pw_gecos = pwdfields[4];
    pwent.pw_dir = pwdfields[5];
    pwent.pw_shell = pwdfields[6];

    return &pwent;
    }

/* fgetpwent - get a password file entry from a stream
**
** fgetpwent() reads the next line from a password file formatted stream
** and returns a pointer to the password structure for that line.
*/
struct passwd*
fgetpwent( fp )
    FILE* fp;
    {
    smpw_init();

    if ( fgets( pwdbuf, BUFSIZ, fp ) == (char*) 0 )
	return (struct passwd*) 0;
    pwdbuf[strlen( pwdbuf ) - 1] = '\0';
    return sgetpwent();
    }

/* getpwent - get a password entry from the password file
**
** getpwent() opens the password file, if not already opened, and reads
** a single entry.  NULL is returned if any errors are encountered reading
** the password file.
*/
struct passwd*
getpwent()
    {
    smpw_init();

    if ( pwdfp == (FILE*) 0 && setpwent() < 0 )
	return (struct passwd*) 0;
    return fgetpwent( pwdfp );
    }

/* igetpwuid - locate a password entry for a given uid using the index */
static struct passwd*
igetpwuid( uid, offP )
    uid_t uid;
    off_t *offP;
    {
    struct passwd* pwd;

    *offP = NO_OFF;

    /* Check that the uid index file was opened successfully. */
    if ( pwufd == -1 )
	return (struct passwd*) 0;
    
    /* Use the uid to index into the uid index file. */
    if ( lseek( pwufd, (off_t) uid * sizeof(off_t), L_SET ) < 0 )
	return (struct passwd*) 0;

    /* Read out a seek address. */
    if ( read( pwufd, (char*) offP, sizeof(off_t) ) != sizeof(off_t) )
	return (struct passwd*) 0;

    /* Seek to the address, making sure it's the beginning of a line. */
    if ( *offP == 0 )
	rewind( pwdfp );
    else
	{
	if ( fseek( pwdfp, *offP - 1, 0 ) < 0 )
	    return (struct passwd*) 0;
	if ( getc( pwdfp ) != '\n' )
	    return (struct passwd*) 0;
	}

    /* Read the entry. */
    pwd = fgetpwent( pwdfp );
    if ( pwd == (struct passwd*) 0 )
	return (struct passwd*) 0;

    /* Check that it's the right one. */
    if ( pwd->pw_uid != uid )
	return (struct passwd*) 0;

    /* Ok! */
    return pwd;
    }

/* isetpwuid - set the index entry for a given uid
**
** isetpwuid() takes a uid and a seek address and writes the address
** to the uid index file.
*/
static void
isetpwuid( uid, off )
    uid_t uid;
    off_t off;
    {
    /* Check that the uid index file was opened successfully. */
    if ( pwufd == -1 )
	return;
    
    /* Use the uid to index into the uid index file. */
    if ( lseek( pwufd, (off_t) uid * sizeof(off_t), L_SET ) < 0 )
	return;

    /* Write the seek address. */
    (void) write( pwufd, (char*) &off, sizeof(off_t) );
    }

/* getpwuid - locate a password entry for a given uid
**
** getpwuid() locates a password file entry for the given uid,
** by use of the index if possible, then by nearby search, and
** lastly by linear search.
*/
struct passwd*
getpwuid( uid )
    uid_t uid;
    {
    struct passwd* pwd;
    off_t off;
    char buf[5000];
    int i;

    smpw_init();

    /* Open if necessary. */
    if ( setpwent() < 0 )
	return (struct passwd*) 0;

    /* First try indexed access. */
    pwd = igetpwuid( uid, &off );
    if ( pwd != (struct passwd*) 0 )
	return pwd;

    if ( off != NO_OFF )
	{
	/* Fall back on nearby search.  Seek backwards a few entries. */
	if ( off < NEARBY_ENTRIES * BYTES_PER_ENTRY / 2 )
	    off = 0;
	else
	    off -= NEARBY_ENTRIES * BYTES_PER_ENTRY / 2;
	if ( fseek( pwdfp, off, 0 ) >= 0 )
	    {
	    if ( off != 0 )
		/* Start at next newline. */
		(void) fgets( buf, sizeof(buf), pwdfp);
	    for ( i = 0; i < NEARBY_ENTRIES; ++i )
		{
		off = ftell( pwdfp );
		pwd = fgetpwent( pwdfp );
		if ( pwd == (struct passwd*) 0 )
		    break;
		if ( pwd->pw_uid == uid )
		    {
		    isetpwuid( uid, off );
		    return pwd;
		    }
		}
	    }
	}

    /* Fall back on linear search. */
    rewind( pwdfp );
    for (;;)
	{
	off = ftell( pwdfp );
	pwd = fgetpwent( pwdfp );
	if ( pwd == (struct passwd*) 0 )
	    break;
	if ( pwd->pw_uid == uid )
	    {
	    /* Got it.  Save the seek address for next time. */
	    isetpwuid( uid, off );
	    return pwd;
	    }
	}

    /* Couldn't find it anywhere. */
    isetpwuid( uid, NO_OFF );
    return (struct passwd*) 0;
    }

/* namehash - hash a name */
static unsigned long
namehash( name )
    char* name;
    {
    char* cp;
    unsigned long hash;

    for ( cp = name, hash = 0; *cp != '\0'; ++cp )
	hash = ( hash * NAMEHASHMULT + *cp ) % namehashsize;

    return hash;
    }

/* isetuident - set the index entry for a given name hash
**
** isetuident() takes a hash entry and a uid and writes the uid
** to the name index file.
*/
static void
isetuident( ent, uid )
    unsigned long ent;
    uid_t uid;
    {
    /* Check that the name index file was opened successfully. */
    if ( pwnfd == -1 )
	return;
    
    /* Use the hash entry to index into the name index file. */
    if ( lseek( pwnfd, (off_t) ent * sizeof(uid_t), L_SET ) < 0 )
	return;

    /* Write the uid. */
    (void) write( pwnfd, (char*) &uid, sizeof(uid_t) );
    }

/* igetpwnam - locate a passwd entry for a given name by use of the index */
static struct passwd*
igetpwnam( name )
    char* name;
    {
    uid_t uid;
    unsigned long hash, h;
    int first;
    struct passwd* pwd;

    /* Check that the name index file was opened successfully. */
    if ( pwnfd == -1 )
	return (struct passwd*) 0;
    
    /* Hash the name. */
    hash = namehash( name ); 

    for ( h = hash, first = 1;
	  h != hash || first;
	  h = ( h + 1 ) % namehashsize, first = 0 )
	{
	/* Use the hash to index into the name index file. */
	if ( lseek( pwnfd, (off_t) h * sizeof(uid_t), L_SET ) < 0 )
	    return (struct passwd*) 0;

	/* Read out a uid. */
	if ( read( pwnfd, (char*) &uid, sizeof(uid_t) ) != sizeof(uid_t) )
	    return (struct passwd*) 0;

	/* Has the hash chain run out? */
	if ( uid == NO_UID )
	    return (struct passwd*) 0;

	/* Fetch the passwd entry for this uid. */
	pwd = getpwuid( uid );
	if ( pwd == (struct passwd*) 0 )
	    {
	    /* No uid?  This must be an obsolete entry.  Zero it and
	    ** continue searching.
	    */
	    isetuident( h, NO_UID );
	    continue;
	    }

	/* Check that it's the right one. */
	if ( strcmp( pwd->pw_name, name ) == 0 )
	    return pwd;
	}

    /* The whole name index was searched without a hit!  It must be
    ** trashed.  Zero it.
    */
    (void) ftruncate( pwnfd, 0 );
    (void) endpwent();
    (void) setpwent();
    return (struct passwd*) 0;
    }

/* isetuidnam - store a uid for a given name into the index */
static void
isetuidnam( name, uid )
    char* name;
    uid_t uid;
    {
    unsigned long hash, h;
    int first;
    uid_t juid;


    /* Check that the name index file was opened successfully. */
    if ( pwnfd == -1 )
	return;
    
    /* Hash the name. */
    hash = namehash( name ); 

    for ( h = hash, first = 1;
	  h != hash || first;
	  h = ( h + 1 ) % namehashsize, first = 0 )
	{
	/* Use the hash to index into the name index file. */
	if ( lseek( pwnfd, (off_t) h * sizeof(uid_t), L_SET ) < 0 )
	    return;

	/* Read out a uid. */
	if ( read( pwnfd, (char*) &juid, sizeof(uid_t) ) != sizeof(uid_t) )
	    return;

	/* Is it free? */
	if ( juid == NO_UID )
	    {
	    /* Fill it in.  Doesn't matter if someone else just filled
	    ** in this hole, whoever loses the race this time will just
	    ** try again next time. */
	    isetuident( h, uid );
	    return;
	    }
	}

    /* The whole name index was searched without a hole!  It must be
    ** trashed.  Zero it.
    */
    (void) ftruncate( pwnfd, 0 );
    (void) endpwent();
    (void) setpwent();
    }

/* getpwnam - locate a password entry for a given name
**
** getpwnam() locates a password file entry for the given name,
** by use of the index if possible and by linear search otherwise.
*/
struct passwd*
getpwnam( name )
    char* name;
    {
    struct passwd* pwd;
    off_t off;

    smpw_init();

    /* Open if necessary. */
    if ( setpwent() < 0 )
	return (struct passwd*) 0;

    /* First try indexed access. */
    pwd = igetpwnam( name );
    if ( pwd != (struct passwd*) 0 )
	return pwd;

    /* Fall back on linear search. */
    rewind( pwdfp );
    for (;;)
	{
	off = ftell( pwdfp );
	pwd = fgetpwent( pwdfp );
	if ( pwd == (struct passwd*) 0 )
	    break;
	if ( strcmp( pwd->pw_name, name ) == 0 )
	    {
	    /* Got it.  Save the seek address and uid for next time. */
	    isetpwuid( pwd->pw_uid, off );
	    isetuidnam( name, pwd->pw_uid );
	    return pwd;
	    }
	}

    /* Couldn't find it anywhere. */
    return (struct passwd*) 0;
    }

/* checkpwu - check and adjust the uid index file
**
** Checks that the uid index file is the right size.  If it's too
** big, truncates it, if it's too small, adds NO_OFF entries.
*/
static void
checkpwu()
    {
    struct stat sb;
    off_t goodsize;
    off_t no_off_buf[1000];
    int i;

    if ( fstat( pwufd, &sb ) < 0 )
	{ pwufd = -1; return; }
    goodsize = ( maxuid + 1 ) * sizeof(off_t);
    if ( sb.st_size > goodsize )
	{
	if ( ftruncate( pwufd, goodsize ) < 0 )
	    { pwufd = -1; return; }
	}
    else if ( sb.st_size < goodsize )
	{
	for ( i = sizeof(no_off_buf) / sizeof(off_t) - 1; i >= 0; --i )
	    no_off_buf[i] = NO_OFF;
	if ( lseek( pwufd, sb.st_size, L_SET ) < 0 )
	    { pwufd = -1; return; }
	while ( sb.st_size + sizeof(no_off_buf) < goodsize )
	    {
	    if ( write( pwufd, (char*) no_off_buf, sizeof(no_off_buf) ) !=
		 sizeof(no_off_buf) )
		{ pwufd = -1; return; }
	    sb.st_size += sizeof(no_off_buf);
	    }
	if ( sb.st_size < goodsize )
	    if ( write( pwufd, (char*) no_off_buf, goodsize - sb.st_size ) !=
		 goodsize - sb.st_size )
		{ pwufd = -1; return; }
	}
    }

/* checkpwn - check and adjust the name index file
**
** Checks that the name index file is the right size.  If it's too
** big, truncates it, and if it's either too big or too small, writes
** it full of NO_UID entries.
*/
static void
checkpwn()
    {
    struct stat sb;
    off_t goodsize;
    uid_t no_uid_buf[1000];
    int i;

    if ( fstat( pwnfd, &sb ) < 0 )
	{ pwnfd = -1; return; }
    goodsize = namehashsize * sizeof(uid_t);
    if ( sb.st_size == goodsize )
	return;
    if ( sb.st_size > goodsize )
	{
	if ( ftruncate( pwnfd, goodsize ) < 0 )
	    { pwnfd = -1; return; }
	}
    for ( i = sizeof(no_uid_buf) / sizeof(uid_t) - 1; i >= 0; --i )
	no_uid_buf[i] = NO_UID;
    if ( lseek( pwnfd, (off_t) 0, L_SET ) < 0 )
	{ pwnfd = -1; return; }
    i = 0;
    while ( i + sizeof(no_uid_buf) < goodsize )
	{
	if ( write( pwnfd, (char*) no_uid_buf, sizeof(no_uid_buf) ) !=
	     sizeof(no_uid_buf) )
	    { pwnfd = -1; return; }
	i += sizeof(no_uid_buf);
	}
    if ( i < goodsize )
	if ( write( pwnfd, (char*) no_uid_buf, goodsize - i ) != goodsize - i )
	    { pwnfd = -1; return; }
    }

/* setpwent - open the password file
**
** setpwent() opens the system password file, or rewinds it if it was
** open already.
*/
int
setpwent()
    {
    smpw_init();

    if ( pwdfp == (FILE*) 0 )
	{
	pwdfp = fopen( PWDFILE, "r" );
	if ( pwdfp == (FILE*) 0 )
	    return -1;
	if ( maxuid == 0 )
	    pwufd = pwnfd = -1;
	else
	{
	    pwufd = open( PWDUFILE, O_RDWR );
	    if ( pwufd != -1 )
		checkpwu();
	    pwnfd = open( PWDNFILE, O_RDWR );
	    if ( pwnfd != -1 )
		checkpwn();
	    }
	}
    else
	rewind( pwdfp );

    return 0;
    }

/* endpwent - close a password file
**
** endpwent() closes the password file if open.
*/
int
endpwent()
    {
    int ret;

    smpw_init();

    ret = 0;

    if ( pwdfp != (FILE*) 0 )
	if ( fclose( pwdfp ) != 0 )
	    ret = -1;
    pwdfp = (FILE*) 0;

    if ( pwufd != -1 )
	if ( close( pwufd ) < 0 )
	    ret = -1;
    pwufd = -1;

    if ( pwnfd != -1 )
	if ( close( pwnfd ) < 0 )
	    ret = -1;
    pwnfd = -1;

    return ret;
    }
